#include "pangulu_cuda.h"
void pangulu_cuda_malloc(void **cuda_address,
    size_t size)
{
    GPU_MEMORY+=size;
    cudaMalloc((cuda_address), size);
}

void pangulu_cuda_free(void *cuda_address)
{
    cudaFree(cuda_address);
    return ;
}

void pangulu_cuda_memcpy_host_to_device_value(calculate_type *cuda_address,
                                               calculate_type *cpu_address,
                                               size_t size)
{
    cudaMemcpy(cuda_address,
               cpu_address,
               size * sizeof(calculate_type),
               cudaMemcpyHostToDevice);
}

void pangulu_cuda_memcpy_device_to_host_value(calculate_type *cpu_address,
                                               calculate_type *cuda_address,
                                               size_t size)
{
    cudaMemcpy(cpu_address,
               cuda_address,
               size * sizeof(calculate_type),
               cudaMemcpyDeviceToHost);
}

void pangulu_cuda_memcpy_host_to_device_int(int_t *cuda_address,
                                            int_t *cpu_address,
                                            size_t size)
{
    cudaMemcpy(cuda_address,
               cpu_address,
               size * sizeof(int_t),
               cudaMemcpyHostToDevice);
}

void pangulu_cuda_memcpy_host_to_device_int32(int_32t *cuda_address,
                                                int_32t *cpu_address,
                                                size_t size)
{
    cudaMemcpy(cuda_address,
               cpu_address,
               size * sizeof(int_32t),
               cudaMemcpyHostToDevice);
}

void pangulu_cuda_memcpy_device_to_host_int(int_t *cpu_address,
                                            int_t *cuda_address,
                                            size_t size)
{
    cudaMemcpy(cpu_address,
               cuda_address,
               size * sizeof(int_t),
               cudaMemcpyDeviceToHost);
}

void pangulu_cuda_devicesynchronize()
{
    cudaDeviceSynchronize();
}

void pangulu_cuda_getDevicenum(int_32t *gpu_num)
{
    cudaGetDeviceCount(gpu_num);
}

int_32t pangulu_cuda_setDevice(int_32t gpu_num,
                           int_32t rank)
{
    int_32t usr_id = rank % gpu_num;
    cudaSetDevice(usr_id);
    return usr_id;
}

__global__ void WarpLevel_sflu(int_t n,
                               int_32t *d_nnzU,
                               int_t *d_cscColPtrA,
                               int_t *d_cscRowIdxA,
                               calculate_type *d_cscValueA,
                               int_t *d_cscColPtrL,
                               int_t *d_cscRowIdxL,
                               calculate_type *d_cscValueL,
                               int_t *d_cscColPtrU,
                               int_t *d_cscRowIdxU,
                               calculate_type *d_cscValueU)
{
    int_t warp_id = (blockDim.x * blockIdx.x + threadIdx.x) / 32;
    int_t warp_local_id = threadIdx.x / 32;
    int_t lane_id = threadIdx.x % 32;

    if (warp_id >= n)
        return;

    const int_t colidx = warp_id;

    int_t colA = d_cscColPtrA[colidx];
    int_t colA1 = d_cscColPtrA[colidx + 1];
    int_t colU = d_cscColPtrU[colidx];
    int_t colU1 = d_cscColPtrU[colidx + 1];
    int_t colL = d_cscColPtrL[colidx];
    int_t colL1 = d_cscColPtrL[colidx + 1];
    int_t loadlen = colA1 - colA;

    // use shared memory
    if (loadlen <= SM_LEN_WARPLEV)
    {
        __shared__ int_t s_idxA[WARP_NUM_WARPLU * SM_LEN_WARPLEV];
        __shared__ calculate_type s_valA[WARP_NUM_WARPLU * SM_LEN_WARPLEV];

        int_t *s_idxA_local = &s_idxA[warp_local_id * SM_LEN_WARPLEV];
        calculate_type *s_valA_local = &s_valA[warp_local_id * SM_LEN_WARPLEV];

        for (int_t i = lane_id; i < loadlen; i += WARP_SIZE)
        {
            s_idxA_local[i] = d_cscRowIdxA[colA + i];
            s_valA_local[i] = d_cscValueA[colA + i];
        }

        //step one
        for (int_t i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int_t rowidx = d_cscRowIdxA[i];
            calculate_type value3 = s_valA_local[i - colA];
            //busy-wait until nnzU[rowidx]==0
            do
            {
                __threadfence();
            } while (d_nnzU[rowidx] != -1);

            int_t rowA = d_cscColPtrA[rowidx];
            int_t rowA1 = d_cscColPtrA[rowidx + 1];
            int_t rowU = d_cscColPtrU[rowidx];
            int_t rowU1 = d_cscColPtrU[rowidx + 1];

            for (int_t j = rowA + rowU1 - rowU + lane_id; j < rowA1; j += WARP_SIZE)
            {
                const int_t Lrowindex = d_cscRowIdxA[j];
                const int_t thecolidx = rowidx;
                int_t flag1 = binarySearch(s_idxA_local, 0, loadlen - 1, Lrowindex);
                int_t flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                s_valA_local[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!lane_id)
            {
                atomicSub(&d_nnzU[colidx], 1);
            }
        }

        //step two
        int_t flag5 = colU1 - colU - 1;
        if (s_valA_local[flag5] > ERROR || s_valA_local[flag5] < -ERROR)
        {
        }
        else
        {
            s_valA_local[flag5] = ERROR;
        }
        calculate_type value5 = s_valA_local[flag5];

        d_cscValueL[colL] = 1;
        for (int_t i = colA + colU1 - colU + lane_id; i < colA1; i += WARP_SIZE)
        {
            const int_t Lrowindex = s_idxA_local[i - colA];
            int_t flagL = colL + i - (colA + colU1 - colU) + 1;
            s_valA_local[i - colA] = s_valA_local[i - colA] / value5;
            d_cscValueL[flagL] = s_valA_local[i - colA];
        }

        for (int_t i = lane_id; i < colU1 - colU; i += WARP_SIZE)
        {
            d_cscValueU[colU + i] = s_valA_local[i];
        }

        for (int_t i = lane_id; i < loadlen; i += WARP_SIZE)
        {
            d_cscValueA[i + colA] = s_valA_local[i];
        }

        if (!lane_id)
        {
            atomicSub(&d_nnzU[colidx], 1);
        }
    }
    // do not use shared memory
    else
    {
        //step one
        for (int_t i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int_t rowidx = d_cscRowIdxA[i];
            calculate_type value3 = d_cscValueA[i];
            //busy-wait until nnzU[rowidx]==0
            do
            {
                __threadfence();
            } while (d_nnzU[rowidx] != -1);

            int_t rowA = d_cscColPtrA[rowidx];
            int_t rowA1 = d_cscColPtrA[rowidx + 1];
            int_t rowU = d_cscColPtrU[rowidx];
            int_t rowU1 = d_cscColPtrU[rowidx + 1];

            for (int_t j = rowA + rowU1 - rowU + lane_id; j < rowA1; j += WARP_SIZE)
            {
                const int_t Lrowindex = d_cscRowIdxA[j];
                const int_t thecolidx = rowidx;
                int_t flag1 = binarySearch(d_cscRowIdxA, colA, colA1 - 1, Lrowindex);
                int_t flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                d_cscValueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!lane_id)
            {
                atomicSub(&d_nnzU[colidx], 1);
            }
        }

        for (int_t i = lane_id; i < colU1 - colU; i += WARP_SIZE)
        {
            d_cscValueU[colU + i] = d_cscValueA[colA + i];
        }

        //step two
        int_t flag5 = colA + colU1 - colU - 1;
        if (d_cscValueA[flag5] > ERROR || d_cscValueA[flag5] < -ERROR)
        {
        }
        else
        {
            d_cscValueA[flag5] = ERROR;
        }

        calculate_type value5 = d_cscValueA[flag5];
        d_cscValueL[colL] = 1;
        for (int_t i = colA + colU1 - colU + lane_id; i < colA1; i += WARP_SIZE)
        {
            const int_t Lrowindex = d_cscRowIdxA[i];
            int_t flagL = colL + i - (colA + colU1 - colU) + 1;
            d_cscValueA[i] = d_cscValueA[i] / value5;
            d_cscValueL[flagL] = d_cscValueA[i];
        }

        if (!lane_id)
        {
            atomicSub(&d_nnzU[colidx], 1);
        }
    }
}

__global__ void BlockLevel_sflu_L1(int_t n,
                                   int_32t *d_nnzU,
                                   int_t *d_cscColPtrA,
                                   int_t *d_cscRowIdxA,
                                   calculate_type *d_cscValueA,
                                   int_t *d_cscColPtrL,
                                   int_t *d_cscRowIdxL,
                                   calculate_type *d_cscValueL,
                                   int_t *d_cscColPtrU,
                                   int_t *d_cscRowIdxU,
                                   calculate_type *d_cscValueU)
{
    const int_t colidx = blockIdx.x;

    __shared__ int_32t s_nnzU[1];

    int_t colA = d_cscColPtrA[colidx];
    int_t colA1 = d_cscColPtrA[colidx + 1];
    int_t smemlen = 256;
    int_t len_a = colA1 - colA;

    // use shared memory
    if (len_a <= smemlen)
    {
        __shared__ int_t s_rowidxA[256];
        __shared__ calculate_type s_valueA[256];

        for (int_t i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            s_rowidxA[i] = d_cscRowIdxA[i + colA];
            s_valueA[i] = d_cscValueA[i + colA];
        }

        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int_t colU = d_cscColPtrU[colidx];
        int_t colU1 = d_cscColPtrU[colidx + 1];
        int_t colL = d_cscColPtrL[colidx];
        int_t colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int_t i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int_t rowidx = d_cscRowIdxA[i];
            calculate_type value3 = s_valueA[i - colA];

            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int_t rowA = d_cscColPtrA[rowidx];
            int_t rowA1 = d_cscColPtrA[rowidx + 1];
            int_t rowU = d_cscColPtrU[rowidx];
            int_t rowU1 = d_cscColPtrU[rowidx + 1];

            for (int_t j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int_t Lrowindex = d_cscRowIdxA[j];
                const int_t thecolidx = rowidx;

                int_t flag1 = binarySearch(s_rowidxA, 0, len_a - 1, Lrowindex);
                int_t flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                s_valueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        //step two
        int_t flag5 = colU1 - colU - 1;
        if (s_valueA[flag5] > ERROR || s_valueA[flag5] < -ERROR)
        {
        }
        else
        {
            s_valueA[flag5] = ERROR;
        }
        calculate_type value5 = s_valueA[flag5];

        d_cscValueL[colL] = 1;
        for (int_t i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int_t Lrowindex = s_rowidxA[i - colA];
            int_t flagL = colL + i - (colA + colU1 - colU) + 1;
            s_valueA[i - colA] = s_valueA[i - colA] / value5;
            d_cscValueL[flagL] = s_valueA[i - colA];
        }
        __syncthreads();

        for (int_t i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = s_valueA[i];
        }

        for (int_t i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            d_cscValueA[i + colA] = s_valueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
    //do not use shared memory
    else
    {
        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int_t colU = d_cscColPtrU[colidx];
        int_t colU1 = d_cscColPtrU[colidx + 1];
        int_t colL = d_cscColPtrL[colidx];
        int_t colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int_t i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int_t rowidx = d_cscRowIdxA[i];
            calculate_type value3 = d_cscValueA[i];
            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int_t rowA = d_cscColPtrA[rowidx];
            int_t rowA1 = d_cscColPtrA[rowidx + 1];
            int_t rowU = d_cscColPtrU[rowidx];
            int_t rowU1 = d_cscColPtrU[rowidx + 1];

            for (int_t j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int_t Lrowindex = d_cscRowIdxA[j];
                const int_t thecolidx = rowidx;
                int_t flag1 = binarySearch(d_cscRowIdxA, colA, colA1 - 1, Lrowindex);
                int_t flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                d_cscValueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        for (int_t i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = d_cscValueA[colA + i];
        }

        //step two
        int_t flag5 = colA + colU1 - colU - 1;
        if (d_cscValueA[flag5] > ERROR || d_cscValueA[flag5] < -ERROR)
        {
        }
        else
        {
            d_cscValueA[flag5] = ERROR;
        }
        calculate_type value5 = d_cscValueA[flag5];
        d_cscValueL[colL] = 1;
        
        for (int_t i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int_t Lrowindex = d_cscRowIdxA[i];
            int_t flagL = colL + i - (colA + colU1 - colU) + 1;
            d_cscValueA[i] = d_cscValueA[i] / value5;
            d_cscValueL[flagL] = d_cscValueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
}

__global__ void BlockLevel_sflu_L2(int_t n,
                                   int_32t *d_nnzU,
                                   int_t *d_cscColPtrA,
                                   int_t *d_cscRowIdxA,
                                   calculate_type *d_cscValueA,
                                   int_t *d_cscColPtrL,
                                   int_t *d_cscRowIdxL,
                                   calculate_type *d_cscValueL,
                                   int_t *d_cscColPtrU,
                                   int_t *d_cscRowIdxU,
                                   calculate_type *d_cscValueU)
{
    const int_t colidx = blockIdx.x;

    __shared__ int_32t s_nnzU[1];

    int_t colA = d_cscColPtrA[colidx];
    int_t colA1 = d_cscColPtrA[colidx + 1];
    int_t smemlen = 512;
    int_t len_a = colA1 - colA;

    // use shared memory
    if (len_a <= smemlen)
    {
        __shared__ int_t s_rowidxA[512];
        __shared__ calculate_type s_valueA[512];

        for (int_t i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            s_rowidxA[i] = d_cscRowIdxA[i + colA];
            s_valueA[i] = d_cscValueA[i + colA];
        }

        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int_t colU = d_cscColPtrU[colidx];
        int_t colU1 = d_cscColPtrU[colidx + 1];
        int_t colL = d_cscColPtrL[colidx];
        int_t colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int_t i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int_t rowidx = d_cscRowIdxA[i];
            calculate_type value3 = s_valueA[i - colA];

            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int_t rowA = d_cscColPtrA[rowidx];
            int_t rowA1 = d_cscColPtrA[rowidx + 1];
            int_t rowU = d_cscColPtrU[rowidx];
            int_t rowU1 = d_cscColPtrU[rowidx + 1];

            for (int_t j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int_t Lrowindex = d_cscRowIdxA[j];
                const int_t thecolidx = rowidx;

                int_t flag1 = binarySearch(s_rowidxA, 0, len_a - 1, Lrowindex);
                int_t flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                s_valueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        //step two
        int_t flag5 = colU1 - colU - 1;
        if (s_valueA[flag5] > ERROR || s_valueA[flag5] < -ERROR)
        {
        }
        else
        {
            s_valueA[flag5] = ERROR;
        }
        calculate_type value5 = s_valueA[flag5];

        d_cscValueL[colL] = 1;
        for (int_t i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int_t Lrowindex = s_rowidxA[i - colA];
            int_t flagL = colL + i - (colA + colU1 - colU) + 1;
            s_valueA[i - colA] = s_valueA[i - colA] / value5;
            d_cscValueL[flagL] = s_valueA[i - colA];
        }
        __syncthreads();

        for (int_t i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = s_valueA[i];
        }

        for (int_t i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            d_cscValueA[i + colA] = s_valueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
    //do not use shared memory
    else
    {
        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int_t colU = d_cscColPtrU[colidx];
        int_t colU1 = d_cscColPtrU[colidx + 1];
        int_t colL = d_cscColPtrL[colidx];
        int_t colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int_t i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int_t rowidx = d_cscRowIdxA[i];

            calculate_type value3 = d_cscValueA[i];

            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int_t rowA = d_cscColPtrA[rowidx];
            int_t rowA1 = d_cscColPtrA[rowidx + 1];
            int_t rowU = d_cscColPtrU[rowidx];
            int_t rowU1 = d_cscColPtrU[rowidx + 1];

            for (int_t j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int_t Lrowindex = d_cscRowIdxA[j];
                const int_t thecolidx = rowidx;

                int_t flag1 = binarySearch(d_cscRowIdxA, colA, colA1 - 1, Lrowindex);
                int_t flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                d_cscValueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        for (int_t i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = d_cscValueA[colA + i];
        }

        //step two
        int_t flag5 = colA + colU1 - colU - 1;
        if (d_cscValueA[flag5] > ERROR || d_cscValueA[flag5] < -ERROR)
        {
        }
        else
        {
            d_cscValueA[flag5] = ERROR;
        }
        calculate_type value5 = d_cscValueA[flag5];

        d_cscValueL[colL] = 1;
        for (int_t i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int_t Lrowindex = d_cscRowIdxA[i];
            int_t flagL = colL + i - (colA + colU1 - colU) + 1;
            d_cscValueA[i] = d_cscValueA[i] / value5;
            d_cscValueL[flagL] = d_cscValueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
}

__global__ void BlockLevel_sflu_L3(int_t n,
                                   int_32t *d_nnzU,
                                   int_t *d_cscColPtrA,
                                   int_t *d_cscRowIdxA,
                                   calculate_type *d_cscValueA,
                                   int_t *d_cscColPtrL,
                                   int_t *d_cscRowIdxL,
                                   calculate_type *d_cscValueL,
                                   int_t *d_cscColPtrU,
                                   int_t *d_cscRowIdxU,
                                   calculate_type *d_cscValueU)
{
    const int_t colidx = blockIdx.x;

    __shared__ int_32t s_nnzU[1];

    int_t colA = d_cscColPtrA[colidx];
    int_t colA1 = d_cscColPtrA[colidx + 1];
    int_t smemlen = 1024;
    int_t len_a = colA1 - colA;

    // use shared memory
    if (len_a <= smemlen)
    {
        __shared__ int_t s_rowidxA[1024];
        __shared__ calculate_type s_valueA[1024];

        for (int_t i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            s_rowidxA[i] = d_cscRowIdxA[i + colA];
            s_valueA[i] = d_cscValueA[i + colA];
        }

        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int_t colU = d_cscColPtrU[colidx];
        int_t colU1 = d_cscColPtrU[colidx + 1];
        int_t colL = d_cscColPtrL[colidx];
        int_t colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int_t i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int_t rowidx = d_cscRowIdxA[i];
            calculate_type value3 = s_valueA[i - colA];

            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int_t rowA = d_cscColPtrA[rowidx];
            int_t rowA1 = d_cscColPtrA[rowidx + 1];
            int_t rowU = d_cscColPtrU[rowidx];
            int_t rowU1 = d_cscColPtrU[rowidx + 1];

            for (int_t j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int_t Lrowindex = d_cscRowIdxA[j];
                const int_t thecolidx = rowidx;

                int_t flag1 = binarySearch(s_rowidxA, 0, len_a - 1, Lrowindex);
                int_t flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                s_valueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        //step two
        int_t flag5 = colU1 - colU - 1;
        if (s_valueA[flag5] > ERROR || s_valueA[flag5] < -ERROR)
        {
        }
        else
        {
            s_valueA[flag5] = ERROR;
        }
        calculate_type value5 = s_valueA[flag5];

        d_cscValueL[colL] = 1;
        for (int_t i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int_t Lrowindex = s_rowidxA[i - colA];
            int_t flagL = colL + i - (colA + colU1 - colU) + 1;
            s_valueA[i - colA] = s_valueA[i - colA] / value5;
            d_cscValueL[flagL] = s_valueA[i - colA];
        }
        __syncthreads();

        for (int_t i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = s_valueA[i];
        }

        for (int_t i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            d_cscValueA[i + colA] = s_valueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
    //do not use shared memory
    else
    {
        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int_t colU = d_cscColPtrU[colidx];
        int_t colU1 = d_cscColPtrU[colidx + 1];
        int_t colL = d_cscColPtrL[colidx];
        int_t colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int_t i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int_t rowidx = d_cscRowIdxA[i];

            calculate_type value3 = d_cscValueA[i];

            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int_t rowA = d_cscColPtrA[rowidx];
            int_t rowA1 = d_cscColPtrA[rowidx + 1];
            int_t rowU = d_cscColPtrU[rowidx];
            int_t rowU1 = d_cscColPtrU[rowidx + 1];

            for (int_t j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int_t Lrowindex = d_cscRowIdxA[j];
                const int_t thecolidx = rowidx;

                int_t flag1 = binarySearch(d_cscRowIdxA, colA, colA1 - 1, Lrowindex);
                int_t flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                d_cscValueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        for (int_t i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = d_cscValueA[colA + i];
        }

        //step two
        int_t flag5 = colA + colU1 - colU - 1;
        if (d_cscValueA[flag5] > ERROR || d_cscValueA[flag5] < -ERROR)
        {
        }
        else
        {
            d_cscValueA[flag5] = ERROR;
        }
        calculate_type value5 = d_cscValueA[flag5];

        d_cscValueL[colL] = 1;
        for (int_t i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int_t Lrowindex = d_cscRowIdxA[i];
            int_t flagL = colL + i - (colA + colU1 - colU) + 1;
            d_cscValueA[i] = d_cscValueA[i] / value5;
            d_cscValueL[flagL] = d_cscValueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
}

__global__ void BlockLevel_sflu_L4(int_t n,
                                   int_32t *d_nnzU,
                                   int_t *d_cscColPtrA,
                                   int_t *d_cscRowIdxA,
                                   calculate_type *d_cscValueA,
                                   int_t *d_cscColPtrL,
                                   int_t *d_cscRowIdxL,
                                   calculate_type *d_cscValueL,
                                   int_t *d_cscColPtrU,
                                   int_t *d_cscRowIdxU,
                                   calculate_type *d_cscValueU)
{
    const int_t colidx = blockIdx.x;

    __shared__ int_32t s_nnzU[1];

    int_t colA = d_cscColPtrA[colidx];
    int_t colA1 = d_cscColPtrA[colidx + 1];
    int_t smemlen = 2048;
    int_t len_a = colA1 - colA;

    // use shared memory
    if (len_a <= smemlen)
    {
        __shared__ int_t s_rowidxA[2048];
        __shared__ calculate_type s_valueA[2048];

        for (int_t i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            s_rowidxA[i] = d_cscRowIdxA[i + colA];
            s_valueA[i] = d_cscValueA[i + colA];
        }

        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int_t colU = d_cscColPtrU[colidx];
        int_t colU1 = d_cscColPtrU[colidx + 1];
        int_t colL = d_cscColPtrL[colidx];
        int_t colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int_t i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int_t rowidx = d_cscRowIdxA[i];
            calculate_type value3 = s_valueA[i - colA];

            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int_t rowA = d_cscColPtrA[rowidx];
            int_t rowA1 = d_cscColPtrA[rowidx + 1];
            int_t rowU = d_cscColPtrU[rowidx];
            int_t rowU1 = d_cscColPtrU[rowidx + 1];

            for (int_t j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int_t Lrowindex = d_cscRowIdxA[j];
                const int_t thecolidx = rowidx;

                int_t flag1 = binarySearch(s_rowidxA, 0, len_a - 1, Lrowindex);
                int_t flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                s_valueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        //step two
        int_t flag5 = colU1 - colU - 1;
        if (s_valueA[flag5] > ERROR || s_valueA[flag5] < -ERROR)
        {
        }
        else
        {
            s_valueA[flag5] = ERROR;
        }
        calculate_type value5 = s_valueA[flag5];

        d_cscValueL[colL] = 1;
        for (int_t i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int_t Lrowindex = s_rowidxA[i - colA];
            int_t flagL = colL + i - (colA + colU1 - colU) + 1;
            s_valueA[i - colA] = s_valueA[i - colA] / value5;
            d_cscValueL[flagL] = s_valueA[i - colA];
        }
        __syncthreads();

        for (int_t i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = s_valueA[i];
        }

        for (int_t i = threadIdx.x; i < len_a; i += blockDim.x)
        {
            d_cscValueA[i + colA] = s_valueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
    //do not use shared memory
    else
    {
        if (!threadIdx.x)
        {
            s_nnzU[0] = d_nnzU[colidx];
        }
        __syncthreads();

        int_t colU = d_cscColPtrU[colidx];
        int_t colU1 = d_cscColPtrU[colidx + 1];
        int_t colL = d_cscColPtrL[colidx];
        int_t colL1 = d_cscColPtrL[colidx + 1];

        //step one
        for (int_t i = colA; i < colA + colU1 - colU - 1; i++)
        {
            const int_t rowidx = d_cscRowIdxA[i];

            calculate_type value3 = d_cscValueA[i];

            //busy-wait until nnzU[rowidx]==0
            if (!threadIdx.x)
            {
                do
                {
                    __threadfence_block();
                } while (d_nnzU[rowidx] != -1);
            }
            __syncthreads();

            int_t rowA = d_cscColPtrA[rowidx];
            int_t rowA1 = d_cscColPtrA[rowidx + 1];
            int_t rowU = d_cscColPtrU[rowidx];
            int_t rowU1 = d_cscColPtrU[rowidx + 1];

            for (int_t j = rowA + rowU1 - rowU + threadIdx.x; j < rowA1; j += blockDim.x)
            {
                const int_t Lrowindex = d_cscRowIdxA[j];
                const int_t thecolidx = rowidx;

                int_t flag1 = binarySearch(d_cscRowIdxA, colA, colA1 - 1, Lrowindex);
                int_t flag2 = binarySearch(d_cscRowIdxA, d_cscColPtrA[thecolidx] + d_cscColPtrU[thecolidx + 1] - d_cscColPtrU[thecolidx], d_cscColPtrA[thecolidx + 1] - 1, Lrowindex);

                d_cscValueA[flag1] -= d_cscValueA[flag2] * value3;
            }

            if (!threadIdx.x)
            {
                atomicSub(&s_nnzU[0], 1);
            }
            __syncthreads();
        }

        for (int_t i = threadIdx.x; i < colU1 - colU; i += blockDim.x)
        {
            d_cscValueU[colU + i] = d_cscValueA[colA + i];
        }

        //step two
        int_t flag5 = colA + colU1 - colU - 1;
        if (d_cscValueA[flag5] > ERROR || d_cscValueA[flag5] < -ERROR)
        {
        }
        else
        {
            d_cscValueA[flag5] = ERROR;
        }
        calculate_type value5 = d_cscValueA[flag5];

        d_cscValueL[colL] = 1;
        for (int_t i = colA + colU1 - colU + threadIdx.x; i < colA1; i += blockDim.x)
        {
            const int_t Lrowindex = d_cscRowIdxA[i];
            int_t flagL = colL + i - (colA + colU1 - colU) + 1;
            d_cscValueA[i] = d_cscValueA[i] / value5;
            d_cscValueL[flagL] = d_cscValueA[i];
        }

        if (!threadIdx.x)
        {
            atomicSub(&s_nnzU[0], 1);
            d_nnzU[colidx] = s_nnzU[0];
        }
    }
}

__device__ int_t binarySearch(int_t *ridx, int_t left, int_t right, int_t target)
{
    int_t mid;
    while (left <= right)
    {
        mid = left + (right - left) / 2;
        if (ridx[mid] == target)
        {
            return mid;
        }
        else if (ridx[mid] > target)
        {
            right = mid - 1;
        }
        else
        {
            left = mid + 1;
        }
    }
    return -1;
}

void pangulu_getrf_cuda_kernel(int_t n,
                               int_t nnzA,
                               int_32t *d_nnzU,
                               int_t *A_CUDA_rowpointer,
                               int_t *A_CUDA_columnindex,
                               calculate_type *A_CUDA_value,
                               int_t *L_CUDA_rowpointer,
                               int_t *L_CUDA_columnindex,
                               calculate_type *L_CUDA_value,
                               int_t *U_CUDA_rowpointer,
                               int_t *U_CUDA_columnindex,
                               calculate_type *U_CUDA_value)
{
    int_t nnz_avrg = nnzA / n;

    int_t num_blocks;
    int_t num_threads;

    if (nnz_avrg <= 32)
    {
        num_blocks = ceil((double)n / WARP_NUM_WARPLU);
        num_threads = WARP_NUM_WARPLU * WARP_SIZE;
        WarpLevel_sflu<<<num_blocks, num_threads>>>(n, d_nnzU, A_CUDA_rowpointer, A_CUDA_columnindex, A_CUDA_value,
                                                    L_CUDA_rowpointer, L_CUDA_columnindex, L_CUDA_value,
                                                    U_CUDA_rowpointer, U_CUDA_columnindex, U_CUDA_value);
    }
    else if (nnz_avrg > 32 && nnz_avrg <= 96)
    {
        num_blocks = n;
        num_threads = 2 * WARP_SIZE;
        BlockLevel_sflu_L1<<<num_blocks, num_threads>>>(n, d_nnzU, A_CUDA_rowpointer, A_CUDA_columnindex, A_CUDA_value,
                                                        L_CUDA_rowpointer, L_CUDA_columnindex, L_CUDA_value,
                                                        U_CUDA_rowpointer, U_CUDA_columnindex, U_CUDA_value);
    }
    else if (nnz_avrg > 96 && nnz_avrg <= 192)
    {
        num_blocks = n;
        num_threads = 4 * WARP_SIZE;
        BlockLevel_sflu_L2<<<num_blocks, num_threads>>>(n, d_nnzU, A_CUDA_rowpointer, A_CUDA_columnindex, A_CUDA_value,
                                                        L_CUDA_rowpointer, L_CUDA_columnindex, L_CUDA_value,
                                                        U_CUDA_rowpointer, U_CUDA_columnindex, U_CUDA_value);
    }
    else if (nnz_avrg > 192 && nnz_avrg <= 384)
    {
        num_blocks = n;
        num_threads = 8 * WARP_SIZE;
        BlockLevel_sflu_L3<<<num_blocks, num_threads>>>(n, d_nnzU, A_CUDA_rowpointer, A_CUDA_columnindex, A_CUDA_value,
                                                        L_CUDA_rowpointer, L_CUDA_columnindex, L_CUDA_value,
                                                        U_CUDA_rowpointer, U_CUDA_columnindex, U_CUDA_value);
    }
    else
    {
        num_blocks = n;
        num_threads = 16 * WARP_SIZE;
        BlockLevel_sflu_L4<<<num_blocks, num_threads>>>(n, d_nnzU, A_CUDA_rowpointer, A_CUDA_columnindex, A_CUDA_value,
                                                        L_CUDA_rowpointer, L_CUDA_columnindex, L_CUDA_value,
                                                        U_CUDA_rowpointer, U_CUDA_columnindex, U_CUDA_value);
    }
}

__global__ void trans_cuda(int_t n, calculate_type *d_val_csr, int_t *d_idx, calculate_type *d_val_csc)
{

    if (blockDim.x * blockIdx.x + threadIdx.x >= n)
        return;

    int_t i = blockDim.x * blockIdx.x + threadIdx.x;
    d_val_csc[d_idx[i]] = d_val_csr[i];
}

void pangulu_cuda_transport_kernel(int_t n, calculate_type *d_val_csr, int_t *d_idx, calculate_type *d_val_csc)
{

    int_t num_threads = WARP_SIZE * 2;
    int_t num_blocks = ceil((double)n / num_threads);

    trans_cuda<<<num_blocks, num_threads>>>(n, d_val_csr, d_idx, d_val_csc);
    cudaDeviceSynchronize();
}

__global__ void vector_add_cuda(int_t nnz, calculate_type *d_now_val_csc, calculate_type *d_old_val_csc)
{

    if (blockDim.x * blockIdx.x + threadIdx.x >= nnz)
        return;

    int_t i = blockDim.x * blockIdx.x + threadIdx.x;
    d_now_val_csc[i] += d_old_val_csc[i];
}

void pangulu_cuda_vector_add_kernel(int_t nnz, calculate_type *d_now_val_csc, calculate_type *d_old_val_csc)
{

    int_t num_threads = WARP_SIZE * 2;
    int_t num_blocks = ceil((double)nnz / num_threads);
    cudaDeviceSynchronize();
    vector_add_cuda<<<num_blocks, num_threads>>>(nnz, d_now_val_csc, d_old_val_csc);
    cudaDeviceSynchronize();
}



__global__ void WarpLevel_spgemm_32(int_t n,
                                    int_t layer,
                                    int_t *d_bin_rowpointer,
                                    int_t *d_bin_rowindex,
                                    int_t *d_rowPtrA,
                                    int_t *d_colIdxA,
                                    calculate_type *d_valueA,
                                    int_t *d_rowPtrB,
                                    int_t *d_colIdxB,
                                    calculate_type *d_valueB,
                                    int_t *d_rowPtrC,
                                    int_t *d_colIdxC,
                                    calculate_type *d_valueC)
{
    int_t warp_global_id = (blockDim.x * blockIdx.x + threadIdx.x) / 32;
    int_t warp_local_id = threadIdx.x / 32;
    int_t lane_id = threadIdx.x % 32;

    if (warp_global_id >= d_bin_rowpointer[layer + 1] - d_bin_rowpointer[layer])
        return;

    const int_t rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + warp_global_id];

    int_t therowc = d_rowPtrC[rowidx];
    int_t nextrowc = d_rowPtrC[rowidx + 1];

    int_t therowa = d_rowPtrA[rowidx];
    int_t nextrowa = d_rowPtrA[rowidx + 1];

    int_t SM_LEN = 32;

    __shared__ int_t s_idxC[32 * WARP_PER_BLOCK_GEMM];
    __shared__ calculate_type s_valC[32 * WARP_PER_BLOCK_GEMM];

    int_t *s_idxC_local = &s_idxC[warp_local_id * SM_LEN];
    calculate_type *s_valC_local = &s_valC[warp_local_id * SM_LEN];

    if (lane_id < nextrowc - therowc)
    {
        s_idxC_local[lane_id] = d_colIdxC[therowc + lane_id];
        s_valC_local[lane_id] = 0;
    }

    for (int_t i = therowa + lane_id; i < nextrowa; i += WARP_SIZE)
    {
        int_t colA = d_colIdxA[i];
        calculate_type valA = d_valueA[i];

        int_t therowb = d_rowPtrB[colA];
        int_t nextrowb = d_rowPtrB[colA + 1];

        for (int_t j = therowb; j < nextrowb; j++)
        {
            int_t colB = d_colIdxB[j];
            int_t flag = binarySearch(s_idxC_local, 0, nextrowc - therowc - 1, colB);
            atomicAdd(&s_valC_local[flag], -valA * d_valueB[j]);
        }
    }

    if (lane_id < nextrowc - therowc)
    {
        d_valueC[therowc + lane_id] += s_valC_local[lane_id];
    }
}

__global__ void WarpLevel_spgemm_64(int_t n,
                                    int_t layer,
                                    int_t *d_bin_rowpointer,
                                    int_t *d_bin_rowindex,
                                    int_t *d_rowPtrA,
                                    int_t *d_colIdxA,
                                    calculate_type *d_valueA,
                                    int_t *d_rowPtrB,
                                    int_t *d_colIdxB,
                                    calculate_type *d_valueB,
                                    int_t *d_rowPtrC,
                                    int_t *d_colIdxC,
                                    calculate_type *d_valueC)
{
    int_t warp_global_id = (blockDim.x * blockIdx.x + threadIdx.x) / 32;
    int_t warp_local_id = threadIdx.x / 32;
    int_t lane_id = threadIdx.x % 32;

    if (warp_global_id >= d_bin_rowpointer[layer + 1] - d_bin_rowpointer[layer])
        return;

    const int_t rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + warp_global_id];

    int_t therowc = d_rowPtrC[rowidx];
    int_t nextrowc = d_rowPtrC[rowidx + 1];

    int_t therowa = d_rowPtrA[rowidx];
    int_t nextrowa = d_rowPtrA[rowidx + 1];

    int_t SM_LEN = 64;

    __shared__ int_t s_idxC[64 * WARP_PER_BLOCK_GEMM];
    __shared__ calculate_type s_valC[64 * WARP_PER_BLOCK_GEMM];

    int_t *s_idxC_local = &s_idxC[warp_local_id * SM_LEN];
    calculate_type *s_valC_local = &s_valC[warp_local_id * SM_LEN];

    for (int_t i = lane_id; i < nextrowc - therowc; i += WARP_SIZE)
    {
        s_idxC_local[i] = d_colIdxC[therowc + i];
        s_valC_local[i] = 0;
    }
    __syncthreads();

    for (int_t i = therowa + lane_id; i < nextrowa; i += WARP_SIZE)
    {
        int_t colA = d_colIdxA[i];
        calculate_type valA = d_valueA[i];

        int_t therowb = d_rowPtrB[colA];
        int_t nextrowb = d_rowPtrB[colA + 1];

        for (int_t j = therowb; j < nextrowb; j++)
        {
            int_t colB = d_colIdxB[j];
            int_t flag = binarySearch(s_idxC_local, 0, nextrowc - therowc - 1, colB);
            atomicAdd(&s_valC_local[flag], -valA * d_valueB[j]);
        }
    }
    for (int_t i = lane_id; i < nextrowc - therowc; i += WARP_SIZE)
    {
        d_valueC[therowc + i] += s_valC_local[i];
    }
    __syncthreads();
}

__global__ void WarpLevel_spgemm_128(int_t n,
                                     int_t layer,
                                     int_t *d_bin_rowpointer,
                                     int_t *d_bin_rowindex,
                                     int_t *d_rowPtrA,
                                     int_t *d_colIdxA,
                                     calculate_type *d_valueA,
                                     int_t *d_rowPtrB,
                                     int_t *d_colIdxB,
                                     calculate_type *d_valueB,
                                     int_t *d_rowPtrC,
                                     int_t *d_colIdxC,
                                     calculate_type *d_valueC)
{
    int_t warp_global_id = (blockDim.x * blockIdx.x + threadIdx.x) / 32;
    int_t warp_local_id = threadIdx.x / 32;
    int_t lane_id = threadIdx.x % 32;

    if (warp_global_id >= d_bin_rowpointer[layer + 1] - d_bin_rowpointer[layer])
        return;

    const int_t rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + warp_global_id];

    int_t therowc = d_rowPtrC[rowidx];
    int_t nextrowc = d_rowPtrC[rowidx + 1];

    int_t therowa = d_rowPtrA[rowidx];
    int_t nextrowa = d_rowPtrA[rowidx + 1];

    int_t SM_LEN = 128;

    __shared__ int_t s_idxC[128 * WARP_PER_BLOCK_GEMM];
    __shared__ calculate_type s_valC[128 * WARP_PER_BLOCK_GEMM];

    int_t *s_idxC_local = &s_idxC[warp_local_id * SM_LEN];
    calculate_type *s_valC_local = &s_valC[warp_local_id * SM_LEN];

    for (int_t i = lane_id; i < nextrowc - therowc; i += WARP_SIZE)
    {
        s_idxC_local[i] = d_colIdxC[therowc + i];
        s_valC_local[i] = 0;
    }
    __syncthreads();

    for (int_t i = therowa; i < nextrowa; i++)
    {
        int_t colA = d_colIdxA[i];
        calculate_type valA = d_valueA[i];

        int_t therowb = d_rowPtrB[colA];
        int_t nextrowb = d_rowPtrB[colA + 1];

        for (int_t j = therowb + lane_id; j < nextrowb; j += WARP_SIZE)
        {
            int_t colB = d_colIdxB[j];
            int_t flag = binarySearch(s_idxC_local, 0, nextrowc - therowc - 1, colB);
            s_valC_local[flag] -= valA * d_valueB[j];
        }
    }

    for (int_t i = lane_id; i < nextrowc - therowc; i += WARP_SIZE)
    {
        d_valueC[therowc + i] += s_valC_local[i];
    }
    __syncthreads();
}

__global__ void ThreadLevel_spgemm(int_t n,
                                   int_t layer,
                                   int_t *d_bin_rowpointer,
                                   int_t *d_bin_rowindex,
                                   int_t *d_rowPtrA,
                                   int_t *d_colIdxA,
                                   calculate_type *d_valueA,
                                   int_t *d_rowPtrB,
                                   int_t *d_colIdxB,
                                   calculate_type *d_valueB,
                                   int_t *d_rowPtrC,
                                   int_t *d_colIdxC,
                                   calculate_type *d_valueC)
{
    const int_t rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + blockDim.x * blockIdx.x + threadIdx.x];

    if (d_bin_rowpointer[layer] + blockDim.x * blockIdx.x + threadIdx.x >= d_bin_rowpointer[layer + 1])
        return;

    int_t therowa = d_rowPtrA[rowidx];
    int_t nextrowa = d_rowPtrA[rowidx + 1];

    int_t therowc = d_rowPtrC[rowidx];

    for (int_t i = therowa; i < nextrowa; i++)
    {
        int_t colA = d_colIdxA[i];
        calculate_type valA = d_valueA[i];

        int_t therowb = d_rowPtrB[colA];
        int_t nextrowb = d_rowPtrB[colA + 1];

        if (nextrowb - therowb != 0)
            d_valueC[therowc] -= valA * d_valueB[therowb];
    }
    __syncthreads();
}

__global__ void BlockLevel_spgemm_256(int_t n,
                                      int_t layer,
                                      int_t *d_bin_rowpointer,
                                      int_t *d_bin_rowindex,
                                      int_t *d_rowPtrA,
                                      int_t *d_colIdxA,
                                      calculate_type *d_valueA,
                                      int_t *d_rowPtrB,
                                      int_t *d_colIdxB,
                                      calculate_type *d_valueB,
                                      int_t *d_rowPtrC,
                                      int_t *d_colIdxC,
                                      calculate_type *d_valueC)
{
    int_t warp_local_id = threadIdx.x / 32;
    int_t warp_num = blockDim.x / 32;
    int_t lane_id = threadIdx.x % 32;

    const int_t rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + blockIdx.x];

    __shared__ int_t s_idxC[256];
    __shared__ calculate_type s_valC[256];

    int_t therowc = d_rowPtrC[rowidx];
    int_t nextrowc = d_rowPtrC[rowidx + 1];

    for (int_t i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        s_idxC[i] = d_colIdxC[therowc + i];
        s_valC[i] = 0;
    }
    __syncthreads();

    int_t therow = d_rowPtrA[rowidx];
    int_t nextrow = d_rowPtrA[rowidx + 1];

    for (int_t i = therow + warp_local_id; i < nextrow; i += warp_num)
    {
        int_t colA = d_colIdxA[i];
        calculate_type valA = d_valueA[i];

        int_t therowb = d_rowPtrB[colA];
        int_t nextrowb = d_rowPtrB[colA + 1];

        for (int_t j = therowb + lane_id; j < nextrowb; j += WARP_SIZE)
        {
            int_t colB = d_colIdxB[j];
            int_t flag = binarySearch(s_idxC, 0, nextrowc - therowc - 1, colB);
            atomicAdd(&s_valC[flag], -valA * d_valueB[j]);
        }
    }
    __syncthreads();

    for (int_t i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        d_valueC[therowc + i] += s_valC[i];
    }
}

__global__ void BlockLevel_spgemm_512(int_t n,
                                      int_t layer,
                                      int_t *d_bin_rowpointer,
                                      int_t *d_bin_rowindex,
                                      int_t *d_rowPtrA,
                                      int_t *d_colIdxA,
                                      calculate_type *d_valueA,
                                      int_t *d_rowPtrB,
                                      int_t *d_colIdxB,
                                      calculate_type *d_valueB,
                                      int_t *d_rowPtrC,
                                      int_t *d_colIdxC,
                                      calculate_type *d_valueC)
{
    int_t warp_local_id = threadIdx.x / 32;
    int_t warp_num = blockDim.x / 32;
    int_t lane_id = threadIdx.x % 32;

    const int_t rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + blockIdx.x];

    __shared__ int_t s_idxC[512];
    __shared__ calculate_type s_valC[512];

    int_t therowc = d_rowPtrC[rowidx];
    int_t nextrowc = d_rowPtrC[rowidx + 1];

    for (int_t i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        s_idxC[i] = d_colIdxC[therowc + i];
        s_valC[i] = 0;
    }
    __syncthreads();

    int_t therow = d_rowPtrA[rowidx];
    int_t nextrow = d_rowPtrA[rowidx + 1];

    for (int_t i = therow + warp_local_id; i < nextrow; i += warp_num)
    {
        int_t colA = d_colIdxA[i];
        calculate_type valA = d_valueA[i];

        int_t therowb = d_rowPtrB[colA];
        int_t nextrowb = d_rowPtrB[colA + 1];

        for (int_t j = therowb + lane_id; j < nextrowb; j += WARP_SIZE)
        {
            int_t colB = d_colIdxB[j];
            int_t flag = binarySearch(s_idxC, 0, nextrowc - therowc - 1, colB);
            atomicAdd(&s_valC[flag], -valA * d_valueB[j]);
        }
    }
    __syncthreads();

    for (int_t i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        d_valueC[therowc + i] += s_valC[i];
    }
}

__global__ void BlockLevel_spgemm_1024(int_t n,
                                       int_t layer,
                                       int_t *d_bin_rowpointer,
                                       int_t *d_bin_rowindex,
                                       int_t *d_rowPtrA,
                                       int_t *d_colIdxA,
                                       calculate_type *d_valueA,
                                       int_t *d_rowPtrB,
                                       int_t *d_colIdxB,
                                       calculate_type *d_valueB,
                                       int_t *d_rowPtrC,
                                       int_t *d_colIdxC,
                                       calculate_type *d_valueC)
{
    int_t warp_local_id = threadIdx.x / 32;
    int_t warp_num = blockDim.x / 32;
    int_t lane_id = threadIdx.x % 32;

    const int_t rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + blockIdx.x];

    __shared__ int_t s_idxC[1024];
    __shared__ calculate_type s_valC[1024];

    int_t therowc = d_rowPtrC[rowidx];
    int_t nextrowc = d_rowPtrC[rowidx + 1];

    for (int_t i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        s_idxC[i] = d_colIdxC[therowc + i];
        s_valC[i] = 0;
    }
    __syncthreads();

    int_t therow = d_rowPtrA[rowidx];
    int_t nextrow = d_rowPtrA[rowidx + 1];

    for (int_t i = therow + warp_local_id; i < nextrow; i += warp_num)
    {
        int_t colA = d_colIdxA[i];
        calculate_type valA = d_valueA[i];

        int_t therowb = d_rowPtrB[colA];
        int_t nextrowb = d_rowPtrB[colA + 1];

        for (int_t j = therowb + lane_id; j < nextrowb; j += WARP_SIZE)
        {
            int_t colB = d_colIdxB[j];
            int_t flag = binarySearch(s_idxC, 0, nextrowc - therowc - 1, colB);
            atomicAdd(&s_valC[flag], -valA * d_valueB[j]);
        }
    }
    __syncthreads();

    for (int_t i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        d_valueC[therowc + i] += s_valC[i];
    }
}

__global__ void BlockLevel_spgemm_2048(int_t n,
                                       int_t layer,
                                       int_t *d_bin_rowpointer,
                                       int_t *d_bin_rowindex,
                                       int_t *d_rowPtrA,
                                       int_t *d_colIdxA,
                                       calculate_type *d_valueA,
                                       int_t *d_rowPtrB,
                                       int_t *d_colIdxB,
                                       calculate_type *d_valueB,
                                       int_t *d_rowPtrC,
                                       int_t *d_colIdxC,
                                       calculate_type *d_valueC)
{
    int_t warp_local_id = threadIdx.x / 32;
    int_t warp_num = blockDim.x / 32;
    int_t lane_id = threadIdx.x % 32;

    const int_t rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + blockIdx.x];

    __shared__ int_t s_idxC[2048];
    __shared__ calculate_type s_valC[2048];

    int_t therowc = d_rowPtrC[rowidx];
    int_t nextrowc = d_rowPtrC[rowidx + 1];

    for (int_t i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        s_idxC[i] = d_colIdxC[therowc + i];
        s_valC[i] = 0;
    }
    __syncthreads();

    int_t therow = d_rowPtrA[rowidx];
    int_t nextrow = d_rowPtrA[rowidx + 1];

    for (int_t i = therow + warp_local_id; i < nextrow; i += warp_num)
    {
        int_t colA = d_colIdxA[i];
        calculate_type valA = d_valueA[i];

        int_t therowb = d_rowPtrB[colA];
        int_t nextrowb = d_rowPtrB[colA + 1];

        for (int_t j = therowb + lane_id; j < nextrowb; j += WARP_SIZE)
        {
            int_t colB = d_colIdxB[j];
            int_t flag = binarySearch(s_idxC, 0, nextrowc - therowc - 1, colB);
            atomicAdd(&s_valC[flag], -valA * d_valueB[j]);
        }
    }
    __syncthreads();

    for (int_t i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
    {
        d_valueC[therowc + i] += s_valC[i];
    }
}

// __global__ void BlockLevel_spgemm_4096(int_t n,
//                                        int_t layer,
//                                        int_t *d_bin_rowpointer,
//                                        int_t *d_bin_rowindex,
//                                        int_t *d_rowPtrA,
//                                        int_t *d_colIdxA,
//                                        calculate_type *d_valueA,
//                                        int_t *d_rowPtrB,
//                                        int_t *d_colIdxB,
//                                        calculate_type *d_valueB,
//                                        int_t *d_rowPtrC,
//                                        int_t *d_colIdxC,
//                                        calculate_type *d_valueC)
// {
//     int_t warp_local_id = threadIdx.x / 32;
//     int_t warp_num = blockDim.x / 32;
//     int_t lane_id = threadIdx.x % 32;

//     const int_t rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + blockIdx.x];

//     __shared__ int_t s_idxC[4096];
//     __shared__ calculate_type s_valC[4096];

//     int_t therowc = d_rowPtrC[rowidx];
//     int_t nextrowc = d_rowPtrC[rowidx + 1];

//     for (int_t i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
//     {
//         s_idxC[i] = d_colIdxC[therowc + i];
//         s_valC[i] = 0;
//     }
//     __syncthreads();

//     int_t therow = d_rowPtrA[rowidx];
//     int_t nextrow = d_rowPtrA[rowidx + 1];

//     for (int_t i = therow + warp_local_id; i < nextrow; i += warp_num)
//     {
//         int_t colA = d_colIdxA[i];
//         calculate_type valA = d_valueA[i];

//         int_t therowb = d_rowPtrB[colA];
//         int_t nextrowb = d_rowPtrB[colA + 1];

//         for (int_t j = therowb + lane_id; j < nextrowb; j += WARP_SIZE)
//         {
//             int_t colB = d_colIdxB[j];
//             int_t flag = binarySearch(s_idxC, 0, nextrowc - therowc - 1, colB);
//             atomicAdd(&s_valC[flag], -valA * d_valueB[j]);
//         }
//     }
//     __syncthreads();

//     for (int_t i = threadIdx.x; i < nextrowc - therowc; i += blockDim.x)
//     {
//         d_valueC[therowc + i] += s_valC[i];
//     }
// }

__global__ void BlockLevel_spgemm_4097(int_t n,
                                       int_t layer,
                                       int_t *d_bin_rowpointer,
                                       int_t *d_bin_rowindex,
                                       int_t *d_rowPtrA,
                                       int_t *d_colIdxA,
                                       calculate_type *d_valueA,
                                       int_t *d_rowPtrB,
                                       int_t *d_colIdxB,
                                       calculate_type *d_valueB,
                                       int_t *d_rowPtrC,
                                       int_t *d_colIdxC,
                                       calculate_type *d_valueC)
{
    int_t warp_local_id = threadIdx.x / 32;
    int_t warp_num = blockDim.x / 32;
    int_t lane_id = threadIdx.x % 32;

    const int_t rowidx = d_bin_rowindex[d_bin_rowpointer[layer] + blockIdx.x];

    int_t therow = d_rowPtrA[rowidx];
    int_t nextrow = d_rowPtrA[rowidx + 1];

    int_t therowc = d_rowPtrC[rowidx];
    int_t nextrowc = d_rowPtrC[rowidx + 1];

    for (int_t i = therow + warp_local_id; i < nextrow; i += warp_num)
    {
        int_t colA = d_colIdxA[i];
        calculate_type valA = d_valueA[i];

        int_t therowb = d_rowPtrB[colA];
        int_t nextrowb = d_rowPtrB[colA + 1];

        for (int_t j = therowb + lane_id; j < nextrowb; j += WARP_SIZE)
        {
            int_t colB = d_colIdxB[j];

            int_t flag = binarySearch(d_colIdxC, therowc, nextrowc - 1, colB);

            atomicAdd(&d_valueC[flag], -valA * d_valueB[j]);
        }

        __syncthreads();
    }
}

void pangulu_ssssm_cuda_kernel(int_t n,
                               int_t *h_bin_rowpointer,
                               int_t *d_bin_rowpointer,
                               int_t *d_bin_rowindex,
                               int_t *d_rowPtrA,
                               int_t *d_colIdxA,
                               calculate_type *d_valueA,
                               int_t *d_rowPtrB,
                               int_t *d_colIdxB,
                               calculate_type *d_valueB,
                               int_t *d_rowPtrC,
                               int_t *d_colIdxC,
                               calculate_type *d_valueC)
{

    int_t num_blocks;
    int_t num_threads;
    int_t layer;

    // 0:0
    layer = 0;

    // 1:1
    layer = 1;
    num_threads = WARP_PER_BLOCK_GEMM * WARP_SIZE;
    num_blocks = ceil((double)(h_bin_rowpointer[2] - h_bin_rowpointer[1]) / num_threads);
    if (num_blocks > 0)
    {
        ThreadLevel_spgemm<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                        d_rowPtrA, d_colIdxA, d_valueA,
                                                        d_rowPtrB, d_colIdxB, d_valueB,
                                                        d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 2:2-32
    layer = 2;
    num_threads = WARP_PER_BLOCK_GEMM * WARP_SIZE;
    num_blocks = ceil((double)(h_bin_rowpointer[3] - h_bin_rowpointer[2]) / WARP_PER_BLOCK_GEMM);
    if (num_blocks > 0)
    {
        WarpLevel_spgemm_32<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                         d_rowPtrA, d_colIdxA, d_valueA,
                                                         d_rowPtrB, d_colIdxB, d_valueB,
                                                         d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 3:33-64
    layer = 3;
    num_threads = WARP_PER_BLOCK_GEMM * WARP_SIZE;
    num_blocks = ceil((double)(h_bin_rowpointer[4] - h_bin_rowpointer[3]) / WARP_PER_BLOCK_GEMM);
    if (num_blocks > 0)
    {
        WarpLevel_spgemm_64<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                         d_rowPtrA, d_colIdxA, d_valueA,
                                                         d_rowPtrB, d_colIdxB, d_valueB,
                                                         d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 4:65-128
    layer = 4;
    num_threads = WARP_PER_BLOCK_GEMM * WARP_SIZE;
    num_blocks = ceil((double)(h_bin_rowpointer[5] - h_bin_rowpointer[4]) / WARP_PER_BLOCK_GEMM);
    if (num_blocks > 0)
    {
        WarpLevel_spgemm_128<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                          d_rowPtrA, d_colIdxA, d_valueA,
                                                          d_rowPtrB, d_colIdxB, d_valueB,
                                                          d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 5:129-256
    layer = 5;
    num_threads = 64;
    num_blocks = h_bin_rowpointer[6] - h_bin_rowpointer[5];
    if (num_blocks > 0)
    {
        BlockLevel_spgemm_256<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                           d_rowPtrA, d_colIdxA, d_valueA,
                                                           d_rowPtrB, d_colIdxB, d_valueB,
                                                           d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 6:257-512
    layer = 6;
    num_threads = 128;
    num_blocks = h_bin_rowpointer[7] - h_bin_rowpointer[6];
    if (num_blocks > 0)
    {
        BlockLevel_spgemm_512<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                           d_rowPtrA, d_colIdxA, d_valueA,
                                                           d_rowPtrB, d_colIdxB, d_valueB,
                                                           d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 7:513-1024
    layer = 7;
    num_threads = 256;
    num_blocks = h_bin_rowpointer[8] - h_bin_rowpointer[7];
    if (num_blocks > 0)
    {
        BlockLevel_spgemm_1024<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                            d_rowPtrA, d_colIdxA, d_valueA,
                                                            d_rowPtrB, d_colIdxB, d_valueB,
                                                            d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 8:1025-2048
    layer = 8;
    num_threads = 512;
    num_blocks = h_bin_rowpointer[9] - h_bin_rowpointer[8];
    if (num_blocks > 0)
    {
        BlockLevel_spgemm_2048<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                            d_rowPtrA, d_colIdxA, d_valueA,
                                                            d_rowPtrB, d_colIdxB, d_valueB,
                                                            d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 9:2049-4096
    layer = 9;
    num_threads = 1024;
    num_blocks = h_bin_rowpointer[10] - h_bin_rowpointer[9];
    if (num_blocks > 0)
    {
        BlockLevel_spgemm_4097<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                            d_rowPtrA, d_colIdxA, d_valueA,
                                                            d_rowPtrB, d_colIdxB, d_valueB,
                                                            d_rowPtrC, d_colIdxC, d_valueC);
    }

    // 10:4097+++
    layer = 10;
    num_threads = 1024;
    num_blocks = h_bin_rowpointer[11] - h_bin_rowpointer[10];
    if (num_blocks > 0)
    {
        BlockLevel_spgemm_4097<<<num_blocks, num_threads>>>(n, layer, d_bin_rowpointer, d_bin_rowindex,
                                                            d_rowPtrA, d_colIdxA, d_valueA,
                                                            d_rowPtrB, d_colIdxB, d_valueB,
                                                            d_rowPtrC, d_colIdxC, d_valueC);
    }
}

__forceinline__ __device__ calculate_type sum_32_shfl(calculate_type sum)
{
#pragma unroll
    for (int_t mask = WARP_SIZE / 2; mask > 0; mask >>= 1)
        sum += __shfl_xor_sync(0xffffffff, sum, mask);

    return sum;
}
__global__ void spmv_warpvec_csr_cuda_executor(const int_t *d_csrRowPtr,
                                               const int_t *d_csrColIdx,
                                               const calculate_type *d_csrVal,
                                               const int_t m,
                                               const calculate_type *d_x,
                                               calculate_type *d_y)
{
    const int_t global_id = blockIdx.x * blockDim.x + threadIdx.x;

    // Initialize
    const int_t rowid = global_id / WARP_SIZE;
    if (rowid >= m)
        return;

    const int_t lane_id = (WARP_SIZE - 1) & threadIdx.x;
    const int_t start = d_csrRowPtr[rowid];
    const int_t stop = d_csrRowPtr[rowid + 1];
    if (start == stop)
    {
        if (!lane_id)
            d_y[rowid] = (calculate_type)0;
        return;
    }

    calculate_type sum = (calculate_type)0;

    for (int_t j = start + lane_id; j < stop; j += WARP_SIZE)
    {
        sum += d_x[d_csrColIdx[j]] * d_csrVal[j];
    }
    sum = sum_32_shfl(sum);

    //finish
    if (!lane_id)
        d_y[rowid] = sum;
}



__global__ void GESSM_Kernel_v2(int_t n,
                                int_t *L_columnpointer,
                                int_t *L_rowindex,
                                calculate_type *L_VALUE,
                                int_t *X_columnpointer,
                                int_t *X_rowindex,
                                calculate_type *X_VALUE,
                                int_t *A_columnpointer,
                                int_t *A_rowindex,
                                calculate_type *A_VALUE)
{
    int_t warp_id = (blockDim.x * blockIdx.x + threadIdx.x) / 32;
    int_t lane_id = threadIdx.x % 32;
    int_t warp_local_id = threadIdx.x / 32;
    int_t flagx, flagb;

    if (warp_id >= n || X_columnpointer[warp_id] == X_columnpointer[warp_id + 1])
        return;

    int_t colA1 = A_columnpointer[warp_id];
    int_t colA2 = A_columnpointer[warp_id + 1];
    int_t colX1 = X_columnpointer[warp_id];
    int_t colX2 = X_columnpointer[warp_id + 1];

    if (colA2 - colA1 >= SM_LEN_DGESSM)
    {
        for (int_t i = colX1, t = 0; i < colX2; i++, t++)
        {
            int_t rowX = X_rowindex[i];
            X_VALUE[i] = A_VALUE[i];
            calculate_type valx = (calculate_type)(X_VALUE[i]);

            int_t colL1 = L_columnpointer[rowX];
            int_t colL2 = L_columnpointer[rowX + 1];
            for (int_t j = colL1 + 1 + lane_id, p = lane_id; j < colL2; j += WARP_SIZE, p += WARP_SIZE)
            {
                int_t f = binarySearch(A_rowindex, colA1 + 1 + t + p, colA2 - colL2 + j, L_rowindex[j]);
                A_VALUE[f] -= valx * L_VALUE[j];
            }
            __syncthreads();
        }
    }
    else
    {
        __shared__ int_t s_rowidxA[SM_LEN_DGESSM * WARP_PER_BLOCK_DGESSM];
        __shared__ calculate_type s_valueA[SM_LEN_DGESSM * WARP_PER_BLOCK_DGESSM];

        int_t *idx_local = &s_rowidxA[warp_local_id * SM_LEN_DGESSM];
        calculate_type *val_local = &s_valueA[warp_local_id * SM_LEN_DGESSM];

        for (int_t i = lane_id; i < colA2 - colA1; i += WARP_SIZE)
        {
            idx_local[i] = A_rowindex[colA1 + i];
            val_local[i] = A_VALUE[colA1 + i];
        }

        for (int_t i = colX1, t = 0; i < colX2; i++, t++)
        {
            int_t rowX = X_rowindex[i];
            X_VALUE[i] = val_local[t];
            calculate_type valx = (calculate_type)(X_VALUE[i]);

            int_t colL1 = L_columnpointer[rowX];
            int_t colL2 = L_columnpointer[rowX + 1];
            for (int_t j = colL1 + 1 + lane_id, p = lane_id; j < colL2; j += WARP_SIZE, p += WARP_SIZE)
            {
                //update A's Value;
                int_t f = binarySearch(idx_local, 1 + t + p, colA2 - colA1 - colL2 + j, L_rowindex[j]);
                if (f != -1)
                    val_local[f] -= valx * L_VALUE[j];
            }
            __syncthreads();
        }

        for (int_t i = lane_id; i < colA2 - colA1; i += WARP_SIZE)
        {
            A_VALUE[colA1 + i] = val_local[i];
        }
    }
}

__global__ void TSTRF_Kernel_v2(int_t n,
                                int_t *U_rowpointer,
                                int_t *U_columnindex,
                                calculate_type *U_VALUE,
                                int_t *X_rowpointer,
                                int_t *X_columnindex,
                                calculate_type *X_VALUE,
                                int_t *A_rowpointer,
                                int_t *A_columnindex,
                                calculate_type *A_VALUE)
{
    int_t warp_id = (blockDim.x * blockIdx.x + threadIdx.x) / 32;
    int_t lane_id = threadIdx.x % 32;
    int_t warp_local_id = threadIdx.x / 32;

    if (warp_id >= n || X_rowpointer[warp_id] == X_rowpointer[warp_id + 1])
        return;

    int_t colA1 = A_rowpointer[warp_id];
    int_t colA2 = A_rowpointer[warp_id + 1];
    int_t colX1 = X_rowpointer[warp_id];
    int_t colX2 = X_rowpointer[warp_id + 1];

    if (colA2 - colA1 >= SM_LEN_DTSTRF)
    {
        for (int_t i = colX1, t = 0; i < colX2; i++, t++)
        {
            int_t rowX = X_columnindex[i];
            int_t colU1 = U_rowpointer[rowX];
            int_t colU2 = U_rowpointer[rowX + 1];

            X_VALUE[i] = A_VALUE[i] / U_VALUE[colU1];
            calculate_type valx = X_VALUE[i];
            for (int_t j = colU1 + 1 + lane_id, p = lane_id; j < colU2; j += WARP_SIZE, p += WARP_SIZE)
            {
                //update A's Value;
                int_t f = binarySearch(A_columnindex, colA1 + t + 1 + p, colA2 - colU2 + j, U_columnindex[j]);
                A_VALUE[f] -= valx * U_VALUE[j];
            }
            __syncthreads();
        }
    }
    else
    {
        __shared__ int_t s_rowidxA[SM_LEN_DTSTRF * WARP_PER_BLOCK_DTSTRF];
        __shared__ calculate_type s_valueA[SM_LEN_DTSTRF * WARP_PER_BLOCK_DTSTRF];

        int_t *idx_local = &s_rowidxA[warp_local_id * SM_LEN_DTSTRF];
        calculate_type *val_local = &s_valueA[warp_local_id * SM_LEN_DTSTRF];

        for (int_t i = lane_id; i < colA2 - colA1; i += WARP_SIZE)
        {
            idx_local[i] = A_columnindex[colA1 + i];
            val_local[i] = A_VALUE[colA1 + i];
        }

        for (int_t i = colX1, t = 0; i < colX2; i++, t++)
        {
            int_t rowX = X_columnindex[i];
            int_t colU1 = U_rowpointer[rowX];
            int_t colU2 = U_rowpointer[rowX + 1];

            X_VALUE[i] = val_local[t] / U_VALUE[colU1];
            calculate_type valx = X_VALUE[i];
            for (int_t j = colU1 + 1 + lane_id, p = lane_id; j < colU2; j += WARP_SIZE, p += WARP_SIZE)
            {
                //update A's Value;
                int_t f = binarySearch(idx_local, 1 + t + p, colA2 - colA1 - colU2 + j, U_columnindex[j]);
                val_local[f] -= valx * U_VALUE[j];
            }
            __syncthreads();
        }

        for (int_t i = lane_id; i < colA2 - colA1; i += WARP_SIZE)
        {
            A_VALUE[colA1 + i] = val_local[i];
        }
    }
}

__global__ void GESSM_Kernel_v3(int_t n,
                                int_t *L_columnpointer,
                                int_t *L_rowindex,
                                calculate_type *L_VALUE,
                                int_t *X_columnpointer,
                                int_t *X_rowindex,
                                calculate_type *X_VALUE,
                                int_t *A_columnpointer,
                                int_t *A_rowindex,
                                calculate_type *A_VALUE)
{

    int_t colidx = blockIdx.x;
    int_t colX1 = X_columnpointer[colidx];
    int_t colX2 = X_columnpointer[colidx + 1];

    if (colidx >= n || colX2 == colX1)
        return;

    int_t colA1 = A_columnpointer[colidx];
    int_t colA2 = A_columnpointer[colidx + 1];

    if (colA2 - colA1 >= SM_LEN_DGESSM)
    {
        for (int_t i = colX1, t = 0; i < colX2; i++, t++)
        {
            int_t rowX = X_rowindex[i];
            X_VALUE[i] = A_VALUE[i];
            calculate_type valx = X_VALUE[i];

            int_t colL1 = L_columnpointer[rowX];
            int_t colL2 = L_columnpointer[rowX + 1];

            for (int_t j = colL1 + 1 + threadIdx.x, p = threadIdx.x; j < colL2; j += blockDim.x, p += blockDim.x)
            {
                //update A's Value;
                int_t f = binarySearch(A_rowindex, colA1 + 1 + t + p, colA2 - colL2 + j, L_rowindex[j]);
                A_VALUE[f] -= valx * L_VALUE[j];
            }
            __syncthreads();
        }
    }
    else
    {
        __shared__ int_t s_idxA[SM_LEN_DGESSM];
        __shared__ calculate_type s_valA[SM_LEN_DGESSM];

        for (int_t i = threadIdx.x; i < colA2 - colA1; i += blockDim.x)
        {
            s_idxA[i] = A_rowindex[colA1 + i];
            s_valA[i] = A_VALUE[colA1 + i];
        }
        __syncthreads();

        for (int_t i = colX1, t = 0; i < colX2; i++, t++)
        {
            int_t rowX = X_rowindex[i];
            X_VALUE[i] = s_valA[t];
            calculate_type valx = X_VALUE[i];

            int_t colL1 = L_columnpointer[rowX];
            int_t colL2 = L_columnpointer[rowX + 1];

            for (int_t j = colL1 + 1 + threadIdx.x, p = threadIdx.x; j < colL2; j += blockDim.x, p += blockDim.x)
            {
                //update A's Value;
                int_t f = binarySearch(s_idxA, 1 + t + p, colA2 - colA1 - colL2 + j, L_rowindex[j]);
                s_valA[f] -= valx * L_VALUE[j];
            }
            __syncthreads();
        }

        for (int_t i = threadIdx.x; i < colA2 - colA1; i += blockDim.x)
        {
            A_VALUE[colA1 + i] = s_valA[i];
        }
        //__syncthreads();
    }
}

__global__ void TSTRF_Kernel_v3(int_t n,
                                int_t *U_rowpointer,
                                int_t *U_columnindex,
                                calculate_type *U_VALUE,
                                int_t *X_rowpointer,
                                int_t *X_columnindex,
                                calculate_type *X_VALUE,
                                int_t *A_rowpointer,
                                int_t *A_columnindex,
                                calculate_type *A_VALUE)
{
    int_t colidx = blockIdx.x;
    int_t colX1 = X_rowpointer[colidx];
    int_t colX2 = X_rowpointer[colidx + 1];

    if (colidx >= n || colX1 == colX2)
        return;

    int_t colA1 = A_rowpointer[colidx];
    int_t colA2 = A_rowpointer[colidx + 1];

    if (colA2 - colA1 >= SM_LEN_DTSTRF)
    {
        for (int_t i = colX1, t = 0; i < colX2; i++, t++)
        {
            int_t rowX = X_columnindex[i];
            int_t colU1 = U_rowpointer[rowX];
            int_t colU2 = U_rowpointer[rowX + 1];

            X_VALUE[i] = A_VALUE[i] / U_VALUE[colU1];
            calculate_type valx = X_VALUE[i];
            for (int_t j = colU1 + 1 + threadIdx.x, p = threadIdx.x; j < colU2; j += blockDim.x, p += blockDim.x)
            {
                //update A's Value;
                int_t f = binarySearch(A_columnindex, colA1 + t + 1 + p, colA2 - colU2 + j, U_columnindex[j]);
                A_VALUE[f] -= valx * U_VALUE[j];
            }
            __syncthreads();
        }
    }
    else
    {
        __shared__ int_t s_rowidxA[SM_LEN_DTSTRF];
        __shared__ calculate_type s_valueA[SM_LEN_DTSTRF];

        for (int_t i = threadIdx.x; i < colA2 - colA1; i += blockDim.x)
        {
            s_rowidxA[i] = A_columnindex[colA1 + i];
            s_valueA[i] = A_VALUE[colA1 + i];
        }
        __syncthreads();

        for (int_t i = colX1, t = 0; i < colX2; i++, t++)
        {
            int_t rowX = X_columnindex[i];
            int_t colU1 = U_rowpointer[rowX];
            int_t colU2 = U_rowpointer[rowX + 1];

            X_VALUE[i] = s_valueA[t] / U_VALUE[colU1];
            calculate_type valx = X_VALUE[i];
            for (int_t j = colU1 + 1 + threadIdx.x, p = threadIdx.x; j < colU2; j += blockDim.x, p += blockDim.x)
            {
                //update A's Value;
                int_t f = binarySearch(s_rowidxA, 1 + t + p, colA2 - colA1 - colU2 + j, U_columnindex[j]);
                s_valueA[f] -= valx * U_VALUE[j];
            }
            __syncthreads();
        }

        for (int_t i = threadIdx.x; i < colA2 - colA1; i += blockDim.x)
        {
            A_VALUE[colA1 + i] = s_valueA[i];
        }
    }
}

void pangulu_gessm_cuda_kernel(int_t n,
                               int_t nnzL,
                               int_t *L_columnpointer,
                               int_t *L_rowindex,
                               calculate_type *L_VALUE,
                               int_t *X_columnpointer,
                               int_t *X_rowindex,
                               calculate_type *X_VALUE,
                               int_t *A_columnpointer,
                               int_t *A_rowindex,
                               calculate_type *A_VALUE)
{
    int_t nnzL_avrg = nnzL / n;

    if (nnzL_avrg < 32)
    {
        int_t num_threads = WARP_PER_BLOCK_DGESSM * WARP_SIZE;
        int_t num_blocks = ceil((double)n / WARP_PER_BLOCK_DGESSM);
        GESSM_Kernel_v2<<<num_blocks, num_threads>>>(n, L_columnpointer, L_rowindex, L_VALUE,
                                                     X_columnpointer, X_rowindex, X_VALUE,
                                                     A_columnpointer, A_rowindex, A_VALUE);
    }
    else
    {
        int_t num_threads = 8 * WARP_SIZE;
        int_t num_blocks = n;
        GESSM_Kernel_v3<<<num_blocks, num_threads>>>(n, L_columnpointer, L_rowindex, L_VALUE,
                                                     X_columnpointer, X_rowindex, X_VALUE,
                                                     A_columnpointer, A_rowindex, A_VALUE);
    }
}

void pangulu_tstrf_cuda_kernel(int_t n,
                               int_t nnzU,
                               int_t *U_rowpointer,
                               int_t *U_columnindex,
                               calculate_type *U_VALUE,
                               int_t *X_rowpointer,
                               int_t *X_columnindex,
                               calculate_type *X_VALUE,
                               int_t *A_rowpointer,
                               int_t *A_columnindex,
                               calculate_type *A_VALUE)
{
    int_t nnzU_avrg = nnzU / n;
    if (nnzU_avrg < 32)
    {
        int_t num_threads = WARP_PER_BLOCK_DTSTRF * WARP_SIZE;
        int_t num_blocks = ceil((double)n / WARP_PER_BLOCK_DTSTRF);
        TSTRF_Kernel_v2<<<num_blocks, num_threads>>>(n, U_rowpointer, U_columnindex, U_VALUE,
                                                     X_rowpointer, X_columnindex, X_VALUE,
                                                     A_rowpointer, A_columnindex, A_VALUE);
    }
    else
    {
        int_t num_threads = 8 * WARP_SIZE;
        int_t num_blocks = n;
        TSTRF_Kernel_v3<<<num_blocks, num_threads>>>(n, U_rowpointer, U_columnindex, U_VALUE,
                                                     X_rowpointer, X_columnindex, X_VALUE,
                                                     A_rowpointer, A_columnindex, A_VALUE);
    }
}
